class DecodingParamsError(Exception):  
    """
    An exception raised when decoding parameters fails.
    """
    pass

class ModelNotMatchingError(Exception):  
    """
    An exception raised when parameters not match with the model.
    """
    pass